@@ -29,6 +29,7 @@
 
 #include "facebook.h"
 #include "facebook-util.h"
+#include "facebook-contact-view.h"
 #include "facebook-item-view.h"
 
 #include <json-glib/json-glib.h>
@@ -46,10 +47,12 @@
 #include <interfaces/sw-status-update-ginterface.h>
 #include <interfaces/sw-photo-upload-ginterface.h>
 #include <interfaces/sw-video-upload-ginterface.h>
+#include <interfaces/sw-contacts-query-ginterface.h>
 #include <interfaces/sw-query-ginterface.h>
 #include <interfaces/sw-collections-ginterface.h>
 
 static void initable_iface_init (gpointer g_iface, gpointer iface_data);
+static void contacts_query_iface_init (gpointer g_iface, gpointer iface_data);
 static void query_iface_init (gpointer g_iface, gpointer iface_data);
 static void avatar_iface_init (gpointer g_iface, gpointer iface_data);
 static void status_update_iface_init (gpointer g_iface, gpointer iface_data);
@@ -62,6 +65,8 @@ G_DEFINE_TYPE_WITH_CODE (SwServiceFacebook,
                          SW_TYPE_SERVICE,
                          G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE,
                                                 initable_iface_init)
+                         G_IMPLEMENT_INTERFACE (SW_TYPE_CONTACTS_QUERY_IFACE,
+                                                contacts_query_iface_init)
                          G_IMPLEMENT_INTERFACE (SW_TYPE_QUERY_IFACE,
                                                 query_iface_init)
                          G_IMPLEMENT_INTERFACE (SW_TYPE_AVATAR_IFACE,
@@ -77,23 +82,6 @@ G_DEFINE_TYPE_WITH_CODE (SwServiceFacebook,
 
 #define GET_PRIVATE(o) (((SwServiceFacebook *) o)->priv)
 
-/* Inspired by services/flickr/flickr.c
-   TODO: This is a useful macro, it should live in lsw */
-#define MAP_PARAM(_call, _fields, _lsw_param, _service_param)      \
-  {                                                                \
-    const char *param = g_hash_table_lookup (_fields, _lsw_param); \
-    if (param)                                                     \
-      rest_proxy_call_add_param (_call, _service_param, param);    \
-  }
-
-#define ALBUM_PREFIX "facebook-"
-
-enum {
-  COLLECTION = 1,
-  PHOTO = 2,
-  VIDEO = 4
-} typedef MediaType;
-
 struct _SwServiceFacebookPrivate {
   gboolean inited;
   gboolean online;
@@ -105,6 +93,31 @@ struct _SwServiceFacebookPrivate {
   char *pic_square;
 };
 
+static const ParameterNameMapping photo_upload_params[] = {
+  { "title", "message" },
+  { NULL, NULL }
+};
+
+static const ParameterNameMapping video_upload_params[] = {
+  { "title", "title" },
+  { "description", "description" },
+  { "x-facebook-privacy", "privacy" },
+  { NULL, NULL }
+};
+
+static const ParameterNameMapping album_create_params[] = {
+  { "x-facebook-privacy", "privacy" },
+  { NULL, NULL }
+}
+  ;
+#define ALBUM_PREFIX "facebook-"
+
+enum {
+  COLLECTION = 1,
+  PHOTO = 2,
+  VIDEO = 4
+} typedef MediaType;
+
 static GList *service_list;
 static const char **
 get_static_caps (SwService *service)
@@ -113,9 +126,11 @@ get_static_caps (SwService *service)
     HAS_UPDATE_STATUS_IFACE,
     HAS_AVATAR_IFACE,
     HAS_BANISHABLE_IFACE,
+    HAS_CONTACTS_QUERY_IFACE,
     HAS_QUERY_IFACE,
     HAS_PHOTO_UPLOAD_IFACE,
     HAS_VIDEO_UPLOAD_IFACE,
+    HAS_COLLECTIONS_IFACE,
 
     /* deprecated */
     CAN_UPDATE_STATUS,
@@ -452,6 +467,42 @@ initable_iface_init (gpointer g_iface, gpointer iface_data)
 }
 
 /* Query interface */
+static void
+_facebook_contacts_query_open_view (SwContactsQueryIface  *self,
+                                    const gchar           *query,
+                                    GHashTable            *params,
+                                    DBusGMethodInvocation *context)
+{
+  SwServiceFacebookPrivate *priv = GET_PRIVATE (self);
+  SwContactView *contact_view;
+  const gchar *object_path;
+
+  g_debug ("query = '%s'", query);
+  if (!g_str_equal (query, "people"))
+  {
+    dbus_g_method_return_error (context,
+                                g_error_new (SW_SERVICE_ERROR,
+                                             SW_SERVICE_ERROR_INVALID_QUERY,
+                                             "Query '%s' is invalid",
+                                             query));
+    return;
+  }
+
+  contact_view = g_object_new (SW_TYPE_FACEBOOK_CONTACT_VIEW,
+                               "service", self,
+                               "proxy", priv->proxy,
+                               "query", query,
+                               "params", params,
+                               NULL);
+  object_path = sw_contact_view_get_object_path (contact_view);
+
+  /* Ensure the object gets disposed when the client goes away */
+  sw_client_monitor_add (dbus_g_method_get_sender (context),
+                         (GObject *) contact_view);
+
+  sw_contacts_query_iface_return_from_open_view (context, object_path);
+}
+
 static void
 _facebook_query_open_view (SwQueryIface          *self,
                            const gchar           *query,
@@ -479,6 +530,14 @@ _facebook_query_open_view (SwQueryIface          *self,
   sw_query_iface_return_from_open_view (context, object_path);
 }
 
+static void
+contacts_query_iface_init (gpointer g_iface, gpointer iface_data)
+{
+  SwContactsQueryIfaceClass *klass = (SwContactsQueryIfaceClass*) g_iface;
+
+  sw_contacts_query_iface_implement_open_view (klass, _facebook_contacts_query_open_view);
+}
+
 static void
 query_iface_init (gpointer g_iface, gpointer iface_data)
 {
@@ -617,6 +676,8 @@ _upload_file (SwServiceFacebook           *self,
     const gchar *album = g_hash_table_lookup (fields, "collection");
     gchar *function;
 
+    call = rest_proxy_new_call (priv->proxy);
+
     if (album != NULL) {
       if (!g_str_has_prefix(album, ALBUM_PREFIX)) {
         g_set_error (error,
@@ -628,16 +689,16 @@ _upload_file (SwServiceFacebook           *self,
       }
 
       function = g_strdup_printf ("%s/photos", album + strlen (ALBUM_PREFIX));
+      rest_proxy_call_set_function (call, function);
+
+      g_free (function);
     } else {
-      function = g_strdup ("me/photos");
+      rest_proxy_call_set_function (call, "me/photos");
     }
 
-    call = rest_proxy_new_call (priv->proxy);
-    rest_proxy_call_set_function (call, function);
-
-    MAP_PARAM (call, fields, "title", "message");
-
-    g_free (function);
+    sw_service_map_params (photo_upload_params, fields,
+                           (SwServiceSetParamFunc) rest_proxy_call_add_param,
+                           call);
   } else if (upload_type == VIDEO ) {
     call = rest_proxy_new_call (priv->video_proxy);
     rest_proxy_call_set_function (call,
@@ -646,9 +707,9 @@ _upload_file (SwServiceFacebook           *self,
                                oauth2_proxy_get_access_token (OAUTH2_PROXY (priv->proxy)));
     rest_proxy_call_add_param (call, "format", "json");
 
-    MAP_PARAM (call, fields, "title", "title");
-    MAP_PARAM (call, fields, "description", "description");
-    MAP_PARAM (call, fields, "x-facebook-privacy", "privacy");
+    sw_service_map_params (video_upload_params, fields,
+                           (SwServiceSetParamFunc) rest_proxy_call_add_param,
+                           call);
   } else {
     g_warning ("invalid upload_type: %d", upload_type);
     goto OUT;
@@ -978,16 +1039,16 @@ _facebook_collections_create (SwCollectionsIface    *self,
 
   if (strlen (collection_parent) != 0) {
     GError error = {SW_SERVICE_ERROR,
-                    SW_SERVICE_ERROR_NOT_SUPPORTED,
-                    "Facebook does not support nested albums."};
+                  SW_SERVICE_ERROR_NOT_SUPPORTED,
+                  "Facebook does not support nested albums."};
     dbus_g_method_return_error (context, &error);
     return;
   }
 
   if (supported_types != PHOTO) {
     GError error = {SW_SERVICE_ERROR,
-                    SW_SERVICE_ERROR_NOT_SUPPORTED,
-                    "Facebook albums can only contain photos."};
+                  SW_SERVICE_ERROR_NOT_SUPPORTED,
+                  "Facebook albums can only contain photos."};
     dbus_g_method_return_error (context, &error);
     return;
   }
@@ -996,7 +1057,9 @@ _facebook_collections_create (SwCollectionsIface    *self,
 
   rest_proxy_call_set_function (call, "me/albums");
 
-  MAP_PARAM (call, extra_fields, "x-facebook-privacy", "privacy");
+  sw_service_map_params (album_create_params, extra_fields,
+                         (SwServiceSetParamFunc) rest_proxy_call_add_param,
+                         call);
 
   rest_proxy_call_add_param (call, "name", collection_name);
   rest_proxy_call_set_method (call, "POST");
@@ -1072,6 +1135,21 @@ _facebook_collections_get_details (SwCollectionsIface    *self,
   g_object_unref (call);
 }
 
+static void
+_facebook_collections_get_creatable_types (SwCollectionsIface    *self,
+                                           DBusGMethodInvocation *context)
+{
+  GArray *creatable_types = g_array_sized_new (TRUE, TRUE,
+                                               sizeof (guint), 1);
+  guint v = PHOTO;
+
+  g_array_append_val (creatable_types, v);
+
+  sw_collections_iface_return_from_get_creatable_types (context,
+                                                        creatable_types);
+
+  g_array_free (creatable_types, TRUE);
+}
 static void
 collections_iface_init (gpointer g_iface,
                         gpointer iface_data)
@@ -1086,4 +1164,7 @@ collections_iface_init (gpointer g_iface,
 
   sw_collections_iface_implement_get_details (klass,
                                               _facebook_collections_get_details);
+
+  sw_collections_iface_implement_get_creatable_types (klass,
+                                                      _facebook_collections_get_creatable_types);
 }
